<?php

declare(strict_types=1);

namespace Drupal\access_by_taxonomy;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\node\NodeInterface;
use Drupal\taxonomy\Entity\Term;

/**
 * Class AccessByTaxonomyService.
 *
 * Contains generic functionality for the module.
 */
final class AccessByTaxonomyService {

  const ROLE_REALM = 'access_by_taxonomy_role';
  const USER_REALM = 'access_by_taxonomy_user';
  const OWNER_REALM = 'access_by_taxonomy_owner';
  const OWN_UNPUBLISHED_REALM = 'access_by_taxonomy_own_unpublished';
  const VIEW_ANY_REALM = 'access_by_taxonomy_view_any_';
  const PUBLIC_REALM = 'access_by_taxonomy_public';

  const ALLOWED_ROLES_FIELD = 'field_allowed_roles';
  const ALLOWED_USERS_FIELD = 'field_allowed_users';

  const DEFAULT_GRANT_VIEW = 1;
  const DEFAULT_GRANT_UPDATE = 0;
  const DEFAULT_GRANT_DELETE = 0;

  /**
   * Constructs an AccessByTaxonomyService object.
   */
  public function __construct(
    private readonly EntityFieldManagerInterface $entityFieldManager,
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly LoggerChannelInterface $loggerChannelAccessByTaxonomy,
    private readonly Connection $connection,
    private readonly AccountInterface $account,
  ) {}

  /**
   * Checks whether there are taxonomy fields defined for a given node type.
   */
  public function getNodeTypeTaxonomyTermFieldDefinitions($nodeType): array {
    $field_definitions = $this->entityFieldManager->getFieldDefinitions('node', $nodeType);
    $return = [];
    foreach ($field_definitions as $field_definition) {
      if ($field_definition->getType() === 'entity_reference' && is_numeric(strpos($field_definition->getSetting('handler'), 'taxonomy_term'))) {
        $return[$field_definition->getName()] = $field_definition->getName();
      }
    }
    return $return;
  }

  /**
   * Determine the grants for a translation of a node.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node translation to examine.
   *
   * @return array
   *   An array of grants.
   */
  public function getTranslationGrants(NodeInterface $node): array {
    $restricted_grants = [];
    // Get the nid from the node.
    $nid = (int) $node->id();
    // Get the langcode of this translation.
    $langcode = $node->language()->getId();
    $authorUserId = (int) $node->getOwnerId();

    // The translation is not published.
    if (!$node->isPublished()) {
      // We add a grant to allow the author of the translation to view it
      // if they have the right permission.
      $restricted_grants[] = $this->createGrantsArray(
        $langcode,
        self::OWN_UNPUBLISHED_REALM,
        $authorUserId,
        self::DEFAULT_GRANT_VIEW,
        self::DEFAULT_GRANT_UPDATE,
        self::DEFAULT_GRANT_DELETE,
        $nid
      );

      // We add a grant specifying the node type, so that users with
      // the 'access by taxonomy view any $type_id content'
      // are able to access the translation even if it's unpublished.
      $realm = self::VIEW_ANY_REALM;
      $realm .= $node->getType();
      $restricted_grants[] = $this->createGrantsArray(
        $langcode,
        $realm,
        1,
        self::DEFAULT_GRANT_VIEW,
        self::DEFAULT_GRANT_UPDATE,
        self::DEFAULT_GRANT_DELETE,
        $nid,
      );
      return $restricted_grants;
    }
    // The node translation is published.
    $role_grants = $this->getRoleGrant();

    // In our module, grants are applied to the node via referenced taxonomy
    // terms. We need to check all the taxonomy fields that the node translation
    // has, and apply the grants accordingly.
    $taxonomy_fields = $this->getNodeTypeTaxonomyTermFieldDefinitions($node->getType());

    $written_grant_users = [];
    $written_grant_roles = [];
    if (!empty($taxonomy_fields)) {
      // The node type has taxonomy fields associated with it, let's examine
      // them:
      foreach ($taxonomy_fields as $field) {
        // Get the values of the field:
        $values = $node->get($field)->getValue();
        foreach ($values as $val) {
          $term = Term::load($val['target_id']);
          if (empty($term)) {
            continue;
          }
          // We check whether any of the terms that the node translation
          // references have allowed roles or users in our special field:
          if ($term->hasField(self::ALLOWED_ROLES_FIELD)) {
            $roles = $term->get(self::ALLOWED_ROLES_FIELD)->getValue();
            foreach ($roles as $role) {
              // Sanity check. The role could be missing from the table.
              if (!isset($role_grants[$role['target_id']])) {
                $this->loggerChannelAccessByTaxonomy->error(
                  'Role :rid not found in access_by_taxonomy_role_gid table.',
                  [
                    ':rid' => $role['target_id'],
                  ]
                );
                continue;
              }
              // Sanity check. Make sure we haven't
              // written this grant already.
              if (isset($written_grant_roles[$role['target_id']])) {
                continue;
              }
              $grant_id = $role_grants[$role['target_id']];
              $written_grant_roles[$role['target_id']] = TRUE;
              $restricted_grants[] = $this->createGrantsArray($langcode,
                self::ROLE_REALM,
                (int) $grant_id,
                self::DEFAULT_GRANT_VIEW,
                self::DEFAULT_GRANT_UPDATE,
                self::DEFAULT_GRANT_DELETE,
                $nid
              );
            }
          }
          // Extract the allowed users from the term
          // and insert grants for them.
          if ($term->hasField(self::ALLOWED_USERS_FIELD)) {
            $users = $term->get(self::ALLOWED_USERS_FIELD)->getValue();
            foreach ($users as $user) {
              // Sanity check.
              // Make sure we haven't written this grant already.
              if (isset($written_grant_users[$user['target_id']])) {
                continue;
              }
              $written_grant_users[$user['target_id']] = TRUE;
              $grant_id = $user['target_id'];
              $restricted_grants[] = $this->createGrantsArray($langcode,
                self::USER_REALM,
                (int) $grant_id,
                self::DEFAULT_GRANT_VIEW,
                self::DEFAULT_GRANT_UPDATE,
                self::DEFAULT_GRANT_DELETE,
                $nid
              );
            }
          }
        }
      }
    }
    // If we have added custom grants for this node, we want to also
    // add a grant for the owner of the node to be able to view it.
    if (count($restricted_grants) > 0) {
      // Make the 'access by taxonomy view any $type_id content'
      // permission work.
      $realm = self::VIEW_ANY_REALM;
      $realm .= $node->getType();
      $restricted_grants[] = $this->createGrantsArray(
        $langcode,
        $realm,
        1,
        self::DEFAULT_GRANT_VIEW,
        self::DEFAULT_GRANT_UPDATE,
        self::DEFAULT_GRANT_DELETE,
        $nid,
      );
    }
    else {
      // The translation is not restricted to any specific role or user.
      // Then we will mark it with our public realm.
      $restricted_grants[] = $this->createGrantsArray(
        $langcode,
        self::PUBLIC_REALM,
        1,
        self::DEFAULT_GRANT_VIEW,
        self::DEFAULT_GRANT_UPDATE,
        self::DEFAULT_GRANT_DELETE,
        $nid,
      );
    }

    // We add a grant to allow the author of the translation to view it.
    $restricted_grants[] = $this->createGrantsArray(
      $langcode,
      self::OWNER_REALM,
      $authorUserId,
      self::DEFAULT_GRANT_VIEW,
      self::DEFAULT_GRANT_UPDATE,
      self::DEFAULT_GRANT_DELETE,
      $nid
    );

    return $restricted_grants;
  }

  /**
   * Creates a grants array.
   */
  public function createGrantsArray(string $langcode, string $realm, int $grant_id, int $grant_view, int $grant_update, int $grant_delete, int $nid): array {
    return [
      'langcode' => $langcode,
      'realm' => $realm,
      'gid' => $grant_id,
      'grant_view' => $grant_view,
      'grant_update' => $grant_update,
      'grant_delete' => $grant_delete,
      'nid' => $nid,
    ];
  }

  /**
   * Gets the node access records for a given node.
   */
  public function getNodeAccessRecords(int $id, string $langcode, string $realm = '*'): array {
    $query = $this->connection->select('node_access');
    $query->fields('node_access', ['nid', 'langcode', 'realm', 'gid']);
    $query->condition('nid', $id);
    $query->condition('realm', $realm);
    $query->condition('langcode', $langcode);
    try {
      return $query->execute()->fetchAll();
    }
    catch (\Exception $e) {
      $this->loggerChannelAccessByTaxonomy->error($e->getMessage());
      return [];
    }
  }

  /**
   * Checks access to a term.
   *
   * @param int $tid
   *   The taxonomy term ID to check for.
   * @param \Drupal\Core\Session\AccountInterface|null $account
   *   The optional account. If not passed, the current user is used.
   *
   * @return bool
   *   Whether the user has access to the term.
   */
  public function canUserAccessTerm(int $tid, ?AccountInterface $account = NULL): bool {

    // If the account is not passed, use the currently logged-in user.
    if (!isset($account)) {
      $account = $this->account;
    }

    $term = Term::load($tid);
    if (!$term) {
      return FALSE;
    }
    $allowed_roles = $term->get(AccessByTaxonomyService::ALLOWED_ROLES_FIELD)->getValue();
    $result = TRUE;
    if ($account->hasPermission('bypass node access')) {
      return $result;
    }
    // In case the term has specified roles, check if the user has any of them.
    if ($allowed_roles) {
      $result = FALSE;
      $user_roles = $account->getRoles();
      $allowed_roles_simplified = [];
      foreach ($allowed_roles as $role) {
        $allowed_roles_simplified[] = $role['target_id'];
      }
      foreach ($user_roles as $role) {
        if (in_array($role, $allowed_roles_simplified)) {
          return TRUE;
        }
      }
    }

    // User has not been allowed by roles, check if in the allowed users list.
    $allowed_users = $term->get(AccessByTaxonomyService::ALLOWED_USERS_FIELD)->getValue();
    if ($allowed_users) {
      $result = FALSE;
      $allowed_users_simplified = [];
      foreach ($allowed_users as $user) {
        $allowed_users_simplified[] = $user['target_id'];
      }
      if (in_array($account->id(), $allowed_users_simplified)) {
        $result = TRUE;
      }
    }

    return $result;

  }

  /**
   * Adds a field to a vocabulary.
   */
  public function addFieldToVocabulary(string $vocabulary, array $field): void {
    $field_name = array_keys($field)[0];
    $field_label = $field[$field_name];

    // Get the storage for the field storage configuration entity:
    $field_config_storage = $this->entityTypeManager->getStorage('field_storage_config');

    // Load field from configuration.
    $field_storage = $field_config_storage->load('taxonomy_term.' . $field_name);
    if ($field_storage) {
      // Define the field.
      $field = FieldConfig::create([
        'field_storage' => $field_storage,
        'field_name' => $field_name,
        'entity_type' => 'taxonomy_term',
        'bundle' => $vocabulary,
        'label' => $field_label,
      ]);
      // Add it to the entity.
      try {
        $field->save();
      }
      catch (EntityStorageException $e) {
        $this->loggerChannelAccessByTaxonomy->error($e->getMessage());
      }

      // Get the storage for the EntityFormDisplay configuration entity:
      $entity_form_display_storage = $this->entityTypeManager->getStorage('entity_form_display');

      // Set the form display.
      // Try to load the existing form display for the taxonomy term bundle:
      $form_display = $entity_form_display_storage->load('taxonomy_term.' . $vocabulary . '.default');

      if (!$form_display) {
        // If the form display does not exist, create it:
        $form_display = $entity_form_display_storage->create([
          'targetEntityType' => 'taxonomy_term',
          'bundle' => $vocabulary,
          'mode' => 'default',
          'status' => TRUE,
        ]);
      }
      if ($field_name == AccessByTaxonomyService::ALLOWED_USERS_FIELD) {
        $form_display->setComponent($field_name, [
          'type' => 'entity_reference_autocomplete',
        ]);
      }
      else {
        $form_display->setComponent($field_name, [
          'type' => 'options_buttons',
        ]);
      }
      try {
        $form_display->save();
      }
      catch (EntityStorageException $e) {
        $this->loggerChannelAccessByTaxonomy->error($e->getMessage());
      }

      // Set the view display.
      $view_display = $entity_form_display_storage->load('taxonomy_term' . $vocabulary . 'default');
      if ($view_display) {
        $view_display->setComponent($field_name, [
          'label' => 'above',
          'type' => 'entity_reference_label',
        ]);
        try {
          $view_display->save();
        }
        catch (EntityStorageException $e) {
          $this->loggerChannelAccessByTaxonomy->error($e->getMessage());
        }
      }
    }
  }

  /**
   * Populates role grants.
   *
   * If the table is not empty, it does not insert anything.
   * Otherwise, it populates the access_by_taxonomy_role_gid table with roles.
   */
  public function populateRoleGrants():void {
    // Sanity check: if the table is not empty, do not insert anything.
    if ($this->getRoleGrant() !== []) {
      return;
    }
    // Populate the access_by_taxonomy_role_gid table.
    $roles = $this->entityTypeManager->getStorage('user_role')->loadMultiple();
    foreach ($roles as $role) {
      $roleId = $role->id();
      $this->insertRoleGrant((string) $roleId);
    }
  }

  /**
   * Gets role grants.
   *
   * @param array $rids
   *   An array of role IDs.
   *
   * @return array
   *   An array of role grants.
   *
   * @throws \Exception
   */
  public function getRoleGrant(array $rids = []) :array {
    $query = $this->connection->select('access_by_taxonomy_role_gid', 't')
      ->fields('t', ['rid', 'gid']);
    if ($rids !== []) {
      $query->condition('t.rid', $rids, 'IN');
    }
    $result = $query->orderBy('t.gid')
      ->execute();
    $grants = [];
    foreach ($result as $grant) {
      $grants[$grant->rid] = $grant->gid;
    }
    return $grants;
  }

  /**
   * Inserts a role grant.
   *
   * @param string $roleId
   *   The role ID.
   *
   * @throws \Exception
   */
  public function insertRoleGrant(string $roleId) : void {
    $fields = [
      'rid' => $roleId,
    ];
    $this->connection->insert('access_by_taxonomy_role_gid')
      ->fields($fields)
      ->execute();
  }

  /**
   * Deletes a role grant.
   *
   * @param string $roleId
   *   The role ID.
   */
  public function deleteRoleGrant(string $roleId) : void {
    $this->connection->delete('access_by_taxonomy_role_gid')
      ->condition('rid', $roleId)
      ->execute();
  }

}
